########################################################
# code author: Dennis Harding
# Class Object: tool for accessing user credentials for GDX
# date 10 Nov 2020
# basic test is in ../tests directory
########################################################

from ckanapi import RemoteCKAN
import os
import sys
import enum
import tempfile
import requests
import shutil


class GdxWarnings(enum.Enum):
    """

    GdxWarnings - enum class of different Gdx warning and errors

    This class provides the enumerated types for reporting warnings, and is
    use to encapsulate the warning into text

    """

    Missing = 1
    Sensitive = 2


class GdxBase:
    """

    GdxBase Class

    GDX Base Class provides the interfaces to download file from the GDX system. From the users perspective they want to be able to download a single resource, or all resources from a data set, to a 'local' directory of their choosing. That directory can be a local disk dir, shared drive, Dropbox or One Drive.

    There are several GDX systems, e.g. test staging, production,...
    GDX is based on the CKAN data catalog tool.

    In theory, users can request one or more data sets, and each dataset can have one or more resources.  when the reply is receives, you can receive one or more packages, each package can have one or more results (datasets), and each result can have one or more resources.
    See the FAQ page at https://dataexchange.gatesfoundation.org/pages/support-index for more
    information.

    To access a GDX system the user must provide a url and an access key. The url points to the CKAN server to use. The access key is unique to the user and system.
    There are three ways to provide the url and api key. First the user can configure a .ckan_profile file at a predefined secure place on their system (see usage set up information in Read,me). This approach allows a user to set up  multiple system and access these system by using user defined service name (see example .ckan_profile).

    The .ckan_profile file contains one line for each system to be accessed. The format of the host definition line is :
    * host_name set_url authtype key

    where ' ' is a 'tab'. lines that begin with a # are considered comments and users can use multiple tabs to line up columns (e.g. multiple 'tabs' are ignored).
    Each host line must have the host name, url and authentication type (authtype) type.
    If the authorization type is 'null', no additional information is required and the rest of the line is ignored.  If the authentication type is 'key', then the key must be provided
    on the line.
    As a future feature we may support other authentication types mat be provided if need.

    Second approach is to store the url and apikey in environment variables.

    :CKAN_API_KEY:  the environment variable containing the  api key
    :CKAN_DEFAULT_URL: the ckan system url

    This approach is good if you are only accessing a single system

    The third approach allows the user to manage the url, and key themselves When creating a GdxBase object if the user does not specify a serviceName,  and does not have environment variables defined, then and empty class is returned and
    the user must specify the server url and access key using the using obj.url and  key by calling the objects url and set_api methods

    GDX contains both public and restricted data, and restricted data is further
    identified as as 'public, restricted, Non-Sensitive, Sensitive and  Very Sensitive.
    Restricted data should only be viewable by Foundation Staff, Contractors, and approved
    Collaborators.  In addition, Sensitive and Very Sensitive data requires a user to agree
    to additional terms and conditions each time the data is downloaded. Currently there is no
    'auto acknowledge' feature  for the API, thus we are not able to download Sensitive
    and Very Sensitive data through the API, only through the Web Application. at this time.

    """

    def __init__(self, serviceName=None):
        """

        class initializer

        What Happens
        When a class is created, the __init__ function is called. if the user supplied a serviceName the ckan_profile is used to  get the service information
        If no service name is specified, but the user has defined  the CKAN_API_KEY an dCKAN_DEFAULT_URL environment variables, the environment variables are used set the service parameters.

        Finally if servicNme url is None, and there are not Environment Variables, a 'empty' class is returned
        If a serviceName url and api key are defined, the connection to the CKAN server is made.

        """

        self.__apiKey = None
        self.__url = None
        self.__tempDir = None

        self.ckanConnection = None
        # warning are a list of tuples consisting 'warnings' and a resource name
        self.warnings = []
        if serviceName is not None and len(serviceName) > 0:    # get url and id from profile
            self._get_profile_data(serviceName)
        elif os.getenv('CKAN_API_KEY') is not None and os.getenv('CKAN_DEFAULT_URL') is not None:
            self.__url = os.getenv('CKAN_DEFAULT_URL')
            self.__apiKey = os.getenv('CKAN_API_KEY')
        if self.__url is not None and len(self.__url) > 0:
            self.connect()

    def set_api_key(self, key):
        """set_api_key

        Allows the user to set the access key for the GDX server

        """

        self.__apiKey = key

    # getter and setter for URL
    def get_url(self):
        return self.__url

    def set_url(self, ckanURL):
        """ set_url


        :param ckanURL: url for the ckan server
        :type ckanURL: str.
        """

        self.__url = ckanURL

    def check_warnings(self):
        """

        Method to check for warnings/errors

        When downloading a data set, because of security issues and system issues
        the requested resource may not be abler to be downloaded. Rather than terminating
        the application because of these issues, the tool logs a warning.
        ant the end of the download, the user should check for warnings.
        :return: list of warnings


        """
        warningText = []
        if len(self.warnings) > 0:
            for warning in self.warnings:
                (wrnType, resource) = warning
                if wrnType == GdxWarnings.Missing:
                    warningText.append("Resource {0} is missing, skipped.".format(resource))
                elif wrnType == GdxWarnings.Sensitive:
                    warningText.append("Resource {0} is sensitive or very sensitive, and requires manual download, skipped.".format(resource))
        return warningText

    def _get_profile_data(self, serviceName):
        """

        Internal routine to get profile data

        get the users profile from the .gdx_profile file.
        Parameters to get are the url and the key. For GDX, the key is required
        For other CKAN systems the key may not be required

        """
        if serviceName is None or serviceName == '':
            raise ValueError("Invalid Service Name Defined")
        # get path to .gdx_profile
        profileFileName = '.ckan_profile'
        windowsAppPaths = ['AppData', 'Roaming', 'CKAN']
        home = os.path.expanduser("~")
        filePath = None

        if sys.platform == 'win32':
            filePath = os.path.join(home, *windowsAppPaths, profileFileName)
        else:
            filePath = os.path.join(home, profileFileName)

        # make sure the profile exists
        if os.path.isfile(filePath):
            self._parse_profile(filePath, serviceName)
        else:
            raise NameError(f"Error: profile file {profileFileName} does not exist")

    def _parse_profile(self, filePath, serviceName):
        """

        Internal routine to parse profile file

        Routine to parse the Profile and extract service information (e.g. url and key)

        """
        with open(filePath, 'r') as file:
            for line in file:
                if len(line) > 0 and line[0] != '#':
                    sys_list = " ".join(line.split()).split()

                    if sys_list[0] == serviceName:  # found the system, get info
                        # There needs to be at least 3 items, system name, url and aurh_type
                        # make sure we have three
                        if len(sys_list) < 3:
                            raise ValueError(f"Definition for system {serviceName} messing required parameters")
                            break

                        self.set_url(sys_list[1])  # url is required
                        auth_type = sys_list[2].lower()
                        if auth_type == 'key' and len(sys_list) > 3:
                            self.set_api_key(sys_list[3])
                        elif auth_type != 'none':
                            self.setId(None)
                        else:
                            # auth_type = 'one'
                            raise ValueError(f"Definition for  {serviceName}, key is missing")
                        break
        # should have a url, may have  a key, otherwise system not found
        if self.__url is None:
            raise NameError(f"Error: Server Name {serviceName} not found in profile")

    def _is_sensitive_data(self, result):
        """

        Internal routine to check if data is Sensitive

        Restricted data can be classified as to sensitivity. This is done by setting
        the resource field information_classification. This routine check for the
        filed, and if present checks to see if the data is sensitive


        """
        sensitive_data = False
        if 'information_classification' in result.keys():
            # check for sensitive and very sensitive data
            classification = result['information_classification']
            if classification.lower() == 'sensitive':
                sensitive_data = True
            elif classification.lower() == 'very_sensitive':
                sensitive_data = True
            else:
                sensitive_data = False
        return sensitive_data

    def _get_output_dir(self, outDir=None):
        if outDir is None:
            self.__tempDir = tempfile.mkdtemp()
            targetDir = self.__tempDir
        else:
            targetDir = outDir
        return targetDir

    def _get_file(self, file, fileUrl, outDir, sensitive, autoAck):
        """Internal Routine to get a file

        This function downloads data from GDX to a file.  If the user supplies an outDir, the file is downloaded to that dir.
        If no output dir is specified, the the file is downloaded to a temporary directory.

        gdx currently does not support and auto acknowledgment feature at this time
        so the auto acknowledge flag (autoAck) is ignored at this time. When auto acknowledge
        is enable, will have to be added when the feature is added to GdxBase.

        """
        targetDir = self._get_output_dir(outDir)
        fileName = None

        fileName = os.path.join(targetDir, file)
        # need to add auto ack to call when available
        try:
            data = requests.get(fileUrl, headers={"Authorization": self.__apiKey})
            open(fileName, 'wb').write(data.content)
        except Exception as e:  # noqa F841

            raise Exception("gdx server error")
        return fileName

    def _download_resource(self, resource, sensitiveData, targetDir, autoAck):
        # if resourceName is specified, then the user wants a specific resource
        dataUrl = resource["url"]

        fileName = dataUrl[(dataUrl.rfind('/') + 1):]
        # at this time we are not allowed to download sensitive or very sensitive
        # data due to security restrictions. once resolved this code will need
        # to change
        file = None
        if not sensitiveData:
            file = self._get_file(fileName, dataUrl, targetDir, sensitiveData, autoAck)
            if file is None:
                self.warnings.append((GdxWarnings.Missing, resource["name"]))
        else:
            self.warnings.append((GdxWarnings.Sensitive, resource["name"]))
        return file

    def _download_dataset_resources(self, results, outDir, autoAck):
        """

        Internal routine _download_dataset_resources

        will download the result, if resultName is None, and a specific
        resource if resource name is specified.


        Args:
            results - is a list of list of results for a gdx package. each result in the
            package request represents a data set, and each dataset can have one or more
            resources


        """
        # get the output directory, if outDir is None, use temporary directory
        targetDir = self._get_output_dir(outDir)

        filesDownloaded = []
        for result in results:
            sensitiveData = self._is_sensitive_data(result)
            for resource in result["resources"]:

                file = self._download_resource(resource, sensitiveData, targetDir, autoAck)
                if file is not None:
                    filesDownloaded.append(file)

        return filesDownloaded

    def cleanup(self):
        # clean up temporary objects.
        if self.__tempDir is not None:
            shutil.rmtree(self.__tempDir)

    def connect(self):
        """connect to ckan server

        Connect to the gdx-server. To connect to the GDX server


        """
        # see if we are already connected
        if self.ckanConnection is not None:
            return True
        # not connected so connect
        if self.__url is not None:
            if self.__apiKey is not None:
                self.ckanConnection = RemoteCKAN(self.__url, apikey=self.__apiKey)
            else:
                self.ckanConnecton = RemoteCKAN(self.__url)
            if self.ckanConnection is not None:
                return True
            else:
                return False
        else:
            raise IOError("service is not connected, check URL and  api Key")
        return False

    def get_group_list(self, groupType):
        """get list of groups by group type

        Args:
             group_type:  string containing group type name

            This function is used to verify you have connected to the GDX or other CKAN server

        """
        groups = None
        if self.ckanConnection:
            groups = self.ckanConnection.action.group_list(id=groupType)
        return groups

    def get_dataset_ids(self, dataset):
        """

        get a list of dataset ids

        Query GDX and get the dataset ID
        Input:
        dataset type string
        output:
        a list list of datasets ids that match the string
        The dataset id is return in the results['id'], it is also available in the
        resource["package_id"]. This routine searches the GDX DB for  for all data sets
        that match the dataset name provided. If multiple results are returned, multiple
        ids will be returned. If no result are returned, and empty list is returned


        Args:
            dataset name

        """
        search_string = 'title: ' + '"' + dataset + '"'
        packages = self.ckanConnection.action.package_search(q=search_string, include_private=True)
        setIds = []
        if packages["results"] is not None:
            for result in packages["results"]:
                if 'id' in result.keys():
                    setIds.append(result["id"])
        return setIds

    def get_dataset_by_name(self, dataset, out_dir=None, auto_ack=False):
        """

        get dataset by name

        get_dataset_by_name: Download all resource assigned to a dataset to a directory out_dir.
        If no directory is given a temporary directory is created. To remove the temporary
        directory, the user must use cleanup() method.
        Note: to down load Sensitive or Very Sensitive data, requires and additional acknowledgment by the user, agreeing that they understand the nature of the data and
        and will will not share the data with others. Currently we do not have a way to auto acknowledge this message. Thus we are not able to download Sensitive or Very Sensitive data at this time. There are discussion to add an auto_ack feature to the API. but until
        the auto_ack feature is defined, the auto_ack parameter will be ignored

        """
        # disable auto_ack
        auto_ack = False
        # reset warnings
        self.warnings = []
        files = None
        if self.ckanConnection is not None:
            search_string = 'title: ' + '"' + dataset + '"'
            packages = self.ckanConnection.action.package_search(q=search_string, include_private=True)
            if packages["results"] is not None:
                files = self._download_dataset_resources(packages["results"], out_dir, auto_ack)
        else:
            raise IOError("service is not connected, check URL and  api Key")
        return files

    def get_dataset_by_id(self, setId, out_dir=None, auto_ack=False):
        """

        get dataset by id

        get_dataset_by_id: Download all resource assigned to a dataset to a directory out_dir.
        If no directory is given a temporary directory is created. To remove the temporary
        directory, the user must use cleanup() method.
        Note: to down load Sensitive or Very Sensitive data, requires and additional acknowledgment by the user, agreeing that they understand the nature of the data and
        and will will not share the data with others. Currently we do not have a way to auto acknowledge this message. Thus we are not able to download Sensitive or Very Sensitive data at this time. There are discussion to add an auto_ack feature to the API. but until
        the auto_ack feature is defined, the auto_ack parameter will be ignored

        """
        # disable auto_ack
        auto_ack = False
        # reset warnings
        self.warnings = []
        files = None
        if self.ckanConnection is not None:
            search_string = 'id:' + '"' + setId + '"'
            packages = self.ckanConnection.action.package_search(q=search_string, include_private=True)
            if packages["results"] is not None:
                files = self._download_dataset_resources(packages["results"], out_dir, auto_ack)
        else:
            raise IOError("service is not connected, check URL and  api Key")

        return files

    def get_resource_by_name(self, dataset, resourceName, out_dir=None, auto_ack=False):
        """get resource by name

        get_resource_by_name get a specific resource from a data set by name

        """
        # reset warnings
        self.warnings = []
        targetDir = self._get_output_dir(out_dir)
        found = False
        file = None
        if self.ckanConnection is not None:
            search_string = 'title: ' + '"' + dataset + '"'
            packages = self.ckanConnection.action.package_search(q=search_string, include_private=True)
            if self.ckanConnection is not None:
                for result in packages["results"]:
                    for resource in result["resources"]:
                        if resource["name"].lower() == resourceName.lower():
                            sensitivity = self._is_sensitive_data(result)
                            file = self._download_resource(resource, sensitivity, targetDir, auto_ack)
                            found = True
                            break
                        if found:
                            break
        else:
            raise IOError("service is not connected, check URL and  api Key")
        if file is None:
            self.warnings.append((GdxWarnings.Missing, resourceName))
        return file

    def get_resource_by_datasetid_and_name(self, id, resourceName, out_dir=None, auto_ack=False):
        """
        get resource by datasetid and name

        get_resource_by_id, will download a resource using the resource id
        Since the resource id is unique, only a single resource is downloaded

        """
        # reset warnings
        self.warnings = []
        targetDir = self._get_output_dir(out_dir)
        found = False
        file = None
        # Create search string
        search_string = 'id:' + '"' + id + '"'
        # Search datasets with this ID (this will only return 1 if your ID is correct)
        file = None

        if self.ckanConnection is not None:
            packages = self.ckanConnection.action.package_search(q=search_string, include_private=True)
            for result in packages["results"]:
                for resource in result["resources"]:
                    if resource["name"].lower() == resourceName.lower():
                        sensitivity = self._is_sensitive_data(result)
                        file = self._download_resource(resource, sensitivity, targetDir, auto_ack)
                        found = True
                        break
                    if found:
                        break
        else:
            raise IOError("service is not connected, check URL and  api Key")
        if file is None:
            self.warnings.append((GdxWarnings.Missing, resourceName))
        return file


